from inspect import cleandoc
from absql.files import parse, accepted_file_types
from absql.files.loader import generate_loader
from jinja2 import Template, DebugUndefined
from absql.functions import default_functions
from absql.text import (
    clean_spacing,
    create_replacements,
    flatten_inputs,
    pretty_encode_sql,
)
from absql.utils import nested_apply, get_function_arg_names, partialize_engine_func


class Runner:
    def __init__(
        self,
        extra_constructors=[],
        replace_only=False,
        file_context_from=None,
        **extra_context,
    ):
        self.extra_context = dict(extra_context)
        self.loader = generate_loader(extra_constructors)
        self.replace_only = replace_only
        self.file_context_from = file_context_from

    @staticmethod
    def render_text(text, replace_only=False, pretty_encode=False, **vars):
        """
        Given some text, render the template with the vars.
        If a templated variable is unknown, leave it alone.
        """

        # Allows an instantiated SQLAlchemy engine to be utilized
        # in any function with a engine argument, without the user needing
        # to specify the engine in the function call.
        engine = vars.get("engine", None)
        for k, v in vars.items():
            if v.__class__.__name__ == "function":
                if "engine" in get_function_arg_names(v):
                    vars[k] = partialize_engine_func(v, engine=engine)

        if replace_only:
            text = clean_spacing(text)
            flat_vars = flatten_inputs(**vars)
            replacements = create_replacements(**flat_vars)
            for k, v in replacements.items():
                text = text.replace(k, str(v))
            text = cleandoc(text)
        else:
            template = Template(text, undefined=DebugUndefined)
            text = cleandoc(template.render(**vars))
        if pretty_encode:
            return pretty_encode_sql(text)
        else:
            return text

    @staticmethod
    def render_context(extra_context=None, file_contents=None):
        """
        Render context dictionaries passed through a function call or
        file frontmatter (file_contents), with file_contents taking
        precedence over other all other provided context.
        """
        rendered_context = default_functions.copy()
        if extra_context:
            rendered_context.update(**extra_context)
        if file_contents:
            rendered_context.update(**file_contents)
        rendered_context = nested_apply(
            rendered_context,
            lambda x: Runner.render_text(x, **rendered_context),
        )
        return rendered_context

    @staticmethod
    def render_file(
        file_path,
        loader=None,
        replace_only=False,
        extra_constructors=[],
        file_context_from=None,
        pretty_encode=False,
        **extra_context,
    ):
        """
        Given a file path, render SQL with a combination of
        the vars in the file and any extras passed to extra_context.
        """
        if loader is None:
            loader = generate_loader(extra_constructors)

        file_contents = parse(file_path, loader=loader)

        sql = file_contents["sql"]
        file_contents.pop("sql")

        if file_context_from:
            file_contents.update(file_contents.get(file_context_from, {}))
            file_contents.pop(file_context_from, {})

        rendered_context = Runner.render_context(extra_context, file_contents)

        rendered = Runner.render_text(
            text=sql,
            replace_only=replace_only,
            pretty_encode=pretty_encode,
            **rendered_context,
        )

        return rendered

    def render(self, text, pretty_encode=False, replace_only=None):
        """
        Given text or a file path, render SQL with the a combination of
        the vars in the file and any extras passed to extra_context during
        the instantiation of the runner.
        """
        if text.endswith(accepted_file_types):
            rendered = self.render_file(
                file_path=text,
                loader=self.loader,
                replace_only=replace_only or self.replace_only,
                file_context_from=self.file_context_from,
                pretty_encode=pretty_encode,
                **self.extra_context,
            )
        else:
            rendered = self.render_text(
                text=text,
                replace_only=replace_only or self.replace_only,
                pretty_encode=pretty_encode,
                **self.render_context(self.extra_context),
            )
        return rendered

    def set_context(self, **context):
        self.extra_context.update(context)



import os
from absql.utils import get_function_arg_names
from absql.files.loader import generate_loader
from absql.files.parsers import parse_generic, parse_sql

default_parsers = {
    ".yml": parse_generic,
    ".yaml": parse_generic,
    ".sql": parse_sql,
}

accepted_file_types = tuple(default_parsers.keys())


def parse(file_path, parse_dict=default_parsers, loader=None):
    """
    Load a file.
    """

    if loader is None:
        loader = generate_loader()

    path, extension = os.path.splitext(file_path)

    parser = parse_dict[extension]

    if "loader" in get_function_arg_names(parser):
        file_parsed = parser(file_path, loader=loader)
    else:
        file_parsed = parser(file_path)

    return file_parsed



import pytest
import mock
import os
from absql import Runner as r


@pytest.fixture(autouse=True)
def mock_settings_env_vars():
    with mock.patch.dict(os.environ, {"name": "Bob"}):
        yield


@pytest.fixture
def extra_sql_path():
    return "tests/files/extra.sql"


@pytest.fixture
def simple_sql_path():
    return "tests/files/simple.sql"


@pytest.fixture
def simple_yml_path():
    return "tests/files/simple.yml"


@pytest.fixture
def constructor_sql_path():
    return "tests/files/constructor.sql"


@pytest.fixture
def constructor_plus_function_sql_path():
    return "tests/files/constructors_plus_functions.sql"


@pytest.fixture
def jinja_frontmatter_path():
    return "tests/files/jinja_frontmatter.sql"


def test_render_simple_sql(simple_sql_path):
    sql = r.render_file(simple_sql_path)
    assert sql == "SELECT * FROM my_table"


def test_render_simple_yml(simple_yml_path):
    sql = r.render_file(simple_yml_path)
    assert sql == "SELECT * FROM my_table"


def test_render_additional_sql(extra_sql_path):
    sql = r.render_file(extra_sql_path, extra="my_extra_context")
    assert sql == "SELECT * FROM my_table WHERE my_extra_context"


def test_render_constructor_sql(constructor_sql_path):
    def add(*nums):
        x = 0
        for n in nums:
            x += n
        return x

    def first(**kwargs):
        for i in range(len(kwargs)):
            if i == 0:
                return kwargs[list(kwargs.keys())[0]]

    def get_table():
        return "my_constructor_table"

    runner = r(extra_constructors=[get_table, add, first])
    sql = runner.render(constructor_sql_path)
    assert sql == "SELECT * FROM my_constructor_table WHERE '6' and 'this'"


def test_render_jinja_frontmatter(jinja_frontmatter_path):
    def provide_table():
        return "my_func_table"

    sql = r.render_file(jinja_frontmatter_path, get_table=provide_table)
    assert sql == "SELECT * FROM my_func_table WHERE name = 'Bob'"


def test_render_jinja_frontmatter_instantiated(jinja_frontmatter_path):
    def provide_table():
        return "my_func_table"

    runner = r(get_table=provide_table)
    sql = runner.render(jinja_frontmatter_path)
    assert sql == "SELECT * FROM my_func_table WHERE name = 'Bob'"


def test_var_dict(extra_sql_path):
    want = "SELECT * FROM my_table WHERE something different!"
    got = r.render_file(extra_sql_path, file_context_from="my_var_dict")
    assert got == want
    runner = r(file_context_from="my_var_dict")
    got = runner.render(extra_sql_path)
    assert got == want


def test_constructor_plus_function(constructor_plus_function_sql_path):
    got = r.render_file(constructor_plus_function_sql_path)
    want = "SELECT * FROM somewhere_else"
    assert got == want



from pendulum import now
from datetime import timedelta


def previous_date(tz="utc", days=1):
    return (now(tz=tz) - timedelta(days=days)).to_date_string()


def previous_hour(tz="utc", hours=1, days=0):
    return (
        (now(tz=tz) - timedelta(hours=hours, days=days))
        .replace(minute=0, second=0)
        .to_datetime_string()
    )



import pytest
from absql import Runner


@pytest.fixture()
def runner():
    return Runner(greeting="Hello")


def test_set_context(runner):
    assert runner.extra_context["greeting"] == "Hello"
    runner.set_context(greeting="Hey")
    assert runner.extra_context["greeting"] == "Hey"



from absql.text import clean_spacing, create_replacements


def test_clean_spaces():
    text = "{{   hello  }}   world"
    got = clean_spacing(text)
    want = "{{ hello }}   world"
    assert got == want


def test_clean_tabs():
    text = "{{  hello   }}   world"
    got = clean_spacing(text)
    want = "{{ hello }}   world"
    assert got == want


def test_replacements():
    got = create_replacements(foo="bar")
    want = {"{{foo}}": "bar", "{{ foo }}": "bar"}
    assert got == want



import os


def env_var(var, default=None):
    return os.environ.get(var, default)


def env_switch(default=None, **kwargs):
    value_unspecified = "value_unspecified"
    if default is None:
        default = value_unspecified
    # What is the environment variable that
    # tells us what environment we are in?
    env_env_var = env_var("AB__ENV", "ENV")
    # What environment are we in?
    env = env_var(env_env_var, value_unspecified)
    # What is the value that corresponds to the environment?
    value = kwargs.get(env, default)
    return value



import re
import yaml
from absql.files.loader import generate_loader


def frontmatter_load(file_path, loader=None):
    """
    Loads YAML frontmatter. Expects a YAML block at the top of the file
    that starts and ends with "---". In use in favor of frontmatter.load
    so that custom dag_constructors (via PyYaml) can be used uniformly across
    all file types.
    """
    if loader is None:
        loader = generate_loader()
    FM_BOUNDARY = re.compile(r"^-{3,}\s*$", re.MULTILINE)
    with open(file_path, "r") as file:
        text = "".join(file.readlines())
        if text.startswith("---"):
            _, metadata, content = FM_BOUNDARY.split(text, 2)
            metadata = yaml.load(metadata, Loader=loader)
            content = content.strip("\n")
        else:
            metadata = {}
            content = yaml.load(text, Loader=loader)
    return {"metadata": metadata, "content": content}


def parse_generic(file_path, loader=None):
    if loader is None:
        loader = generate_loader()
    raw_content = frontmatter_load(file_path)
    file_content = raw_content["metadata"] or raw_content["content"]
    return file_content


def parse_sql(file_path, loader=None):
    if loader is None:
        loader = generate_loader()
    raw_content = frontmatter_load(file_path, loader=loader)
    file_content = raw_content["metadata"]
    file_content["sql"] = raw_content["content"]
    return file_content
